/*  batt_warn -- Acoustically warn the user when the battery is running out
 *  Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
 *  
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 *  02110-1301, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#define _GNU_SOURCE
#define __USE_GNU
#include <string.h>

struct actions {
    long int time;
    const char *action;
};

/* 
 * Config:
 *   MINIMUM: How many seconds before power failure the acoustic warnings
 *            should start (in seconds)
 *   INTERVAL: Minimum interval when remaining time should be checked. If
 *            there's enough time left larger intervals will be taken anyway.
 *   PLAYER: Player to use to play the sounds
 *   BASE_PATH: Where to search for the sounds
 *
 *   sounds: sounds.time is the remaining time, sounds.action the filename of
 *           the sound file (relative to BASE_PATH). Longer times must come first!
 *
 *           All INTERVAL seconds the sound that's currently appropriate will
 *           be played.
 *           Example:
 *               1.00*MINIMUM to 0.85*MINIMUM: ding.wav
 *               0.85*MINIMUM to 0.70*MINIMUM: ding02.wav
 *             ...
 *   run_once: The specified `action` will be executed when only `time` seconds
 *             are left. Each entry will be executed only once.
*/

#define MINIMUM (60*20)
#define INTERVAL (60)

static const char PLAYER[] = "/usr/bin/aplay";
static const char BASE_PATH[] = "./";

static const struct actions sounds[] =
{
    { MINIMUM*100/100, "ding.wav" },
    { MINIMUM*85/100, "ding02.wav" },
    { MINIMUM*70/100, "ding05.wav" },
    { MINIMUM*55/100, "ding10.wav" },
    { MINIMUM*40/100, "ding20.wav" },
    { MINIMUM*25/100, "ding40.wav" },
};

static struct actions run_once[] =
{
    { MINIMUM*40/100, "killall cplay; killall vlc; amixer set Master unmute;" },
};

#define nsounds (sizeof (sounds) / sizeof (sounds[0]))
#define nprograms (sizeof (run_once) / sizeof (run_once[0]))

#define TRUE 1
#define FALSE 0

char *parse(char *buf, const char* search) {
    char *location;
    char *end;
    char *pattern = alloca(strlen(search)+3);

    pattern[0] = '\n';
    pattern[1] = '\0';
    strcat(pattern, search);
    strcat(pattern, ":");

    location = strstr(buf, pattern);
    if (!location) {
        fprintf(stderr, "Couldn't find entry '%s'\n", search);
        return NULL;
    }

    location += strlen(pattern);
    while (*location == ' ')
        location++;

    end = strchrnul(location, '\n');
    return strndup(location, end-location);
}

int read_file(char* file, char* buf, size_t bufsize) {
    int fd = open(file, O_RDONLY);
    size_t nbytes;

    if (fd == -1) {
        fprintf(stderr, "Could not open battery state file\n");
        return FALSE;
    }

    nbytes = read(fd, buf, bufsize-1);
    if (nbytes == -1) {
        fprintf(stderr, "Could not read from the battery state file\n");
        return FALSE;
    }
    if (nbytes == bufsize-1) {
        fprintf(stderr, "Buffer too small to read the whole file\n");
        return FALSE;
    }
    buf[nbytes] = '\0';
    close(fd);

    return TRUE;
}

static const long int REM_ERROR = -1;
static const long int REM_NOT_DISCHARGING = -2;

long int calc_remaining(char *buf) {
    long int rate, remaining, result;

    char *srate = parse(buf, "present rate");
    if (!srate)
        return REM_ERROR;

    char *sremaining = parse(buf, "remaining capacity");
    if (!sremaining) {
        free(srate);
        return REM_ERROR;
    }

    if (sscanf(srate, "%ld mA", &rate) != 1) {
        fprintf(stderr, "Failed to parse rate\n");
        result = REM_ERROR;
        goto out;
    }

    if (sscanf(sremaining, "%ld mAh", &remaining) != 1) {
        fprintf(stderr, "Failed to parse rate\n");
        result = REM_ERROR;
        goto out;
    }

    if (remaining < 0 || rate < 0) {
        fprintf(stderr, "rate or remaining < 0. rate=%ld, remaining=%ld\n",
                rate, remaining);
        result = REM_ERROR;
        goto out;
    }

    result = remaining * 3600 / rate;
out:
    free(srate);
    free(sremaining);
    return result;
}

long int remaining() {
    char buf[BUFSIZ];
    char *state;
    long int res;

    if (!read_file("/proc/acpi/battery/BAT1/state", buf, sizeof(buf)))
        return REM_ERROR;

    state = parse(buf, "charging state");
    if (!state)
        return REM_ERROR;

    if (0 != strcmp(state, "discharging")) {
        free(state);
        return REM_NOT_DISCHARGING;
    }

    res = calc_remaining(buf);
    if (res == REM_ERROR)
        return REM_ERROR;

    return res;
}

inline long int max(long int a, long int b) {
    return (a<b) ? b : a;
}

long int calc_sleep_interval(int remain) {
    int sleep_sec;

    sleep_sec = max(remain / 20, INTERVAL);
    if (remain - sleep_sec > MINIMUM)
        return sleep_sec;

    if (remain < MINIMUM)
        return INTERVAL;
    else
        return max(remain - MINIMUM, INTERVAL);

}

void issue_warning(long int remain) {
    int i;
    for (i=0; i < nprograms; i++) {
        if (run_once[i].time >= remain && run_once[i].action) {
            system(run_once[i].action);
            run_once[i].action = NULL;
        }
    }
    for (i=nsounds-1; i >= 0; i--) {
        if (sounds[i].time >= remain) {
            char cmdline[sizeof(PLAYER)+4+sizeof(BASE_PATH)+strlen(sounds[i].action)+3];
            sprintf(cmdline, "%s -q %s%s &", PLAYER, BASE_PATH, sounds[i].action);
            printf("WARNING: LOW BAT. Calling '%s'\n", cmdline);
            system(cmdline);
            break;
        }
    }
}

int main() {
    long int remain;
    long int sleep_sec;
    
    while (TRUE) {
        remain = remaining();
        if (remain == REM_ERROR)
            return 1;
        if (remain == REM_NOT_DISCHARGING)
            break;

        if (remain < MINIMUM) {
            issue_warning(remain);
        }

        printf("Time: %ld\n", remain/60);
        sleep_sec = calc_sleep_interval(remain);
        printf("Sleeping: %ld\n", sleep_sec/60);
        sleep(sleep_sec);
    }

    return 0;
}
